# frozen_string_literal: true

require_relative "error/decision/north_won"
require_relative "error/decision/south_won"

require_relative "error/game_over/captured_king"
require_relative "error/game_over/illegal_move/occupied_square"
require_relative "error/game_over/illegal_move/opponent_piece"
require_relative "error/game_over/move_limit"
require_relative "error/game_over/repetition"
require_relative "error/game_over/time_forfeit"

require_relative "arbiter"
require_relative "position"
require_relative "state"
require_relative "tablebase"

module Nichugi
  class Game
    AUTHORIZED_OCCURRENCE_OF_POSITION = 2

    def initialize(timestamp, max_seconds_per_move:, max_moves: 100, shogi_move_first: true)
      @max_moves            = Integer(max_moves)
      @max_seconds_per_move = Integer(max_seconds_per_move)
      @shogi_move_first     = !shogi_move_first.nil?
      @states               = [State.new(starting_position, Integer(timestamp))]
      @tablebase            = Tablebase.new

      freeze
    end

    def move!(source, target, timestamp)
      is_drop = last_position.class.drop?(source)
      source = last_position.class.drop_index(source) if is_drop

      try!(is_drop, source, target, timestamp)
    rescue Error::GameOver::MoveLimit
      if Arbiter.new(last_position).north_better?
        raise Error::Decision::NorthWon
      else
        raise Error::Decision::SouthWon
      end
    rescue Error::GameOver::CapturedKing, Error::GameOver::IllegalMove, Error::GameOver::Repetition,
           Error::GameOver::TimeForfeit
      if last_position.turn_to_north?
        raise Error::Decision::SouthWon
      else
        raise Error::Decision::NorthWon
      end
    end

    def inspect
      "<Nichugi::Game position ##{@states.length}>"
    end

    private

    def starting_position
      @shogi_move_first ? Position.shogi_vs_xiangqi99 : Position.xiangqi99_vs_shogi
    end

    def try!(is_drop, source, target, timestamp)
      raise Error::GameOver::TimeForfeit if time_forfeit?(timestamp)

      piece = owned(last_position.piece(source, from_hand: is_drop))
      piece = promote(piece) if promote?(source, target)

      content = last_position.square_content(target)

      if is_drop
        raise Error::GameOver::OccupiedSquare unless content.nil?
      else
        content = own(unpromote(content)) unless content.nil?
        possibilities = @tablebase.possibilities(piece: piece, source: source, target: target)
        raise Error::GameOver::IllegalMove unless any?(*possibilities)
      end

      new_position = last_position.after(source, target, piece, content)

      counter = @states.count { |state| state.position.inspect == new_position.inspect }
                       .next

      raise Error::GameOver::Repetition if counter > AUTHORIZED_OCCURRENCE_OF_POSITION

      commit!(new_position, timestamp)

      raise Error::GameOver::CapturedKing if king?(content)
      raise Error::GameOver::MoveLimit if move_limit_reached?
    end

    def commit!(new_position, timestamp)
      @states << State.new(new_position, timestamp)
    end

    def move_limit_reached?
      (@states.length / 2) >= @max_moves
    end

    def king?(content)
      content&.include?("-")
    end

    def any?(*possibilities)
      possibilities.any? { |conditions| last_position.match?(conditions) }
    end

    def promote?(*indexes)
      indexes.any? { |index| last_position.promotion_square?(index) }
    end

    def promotable_pieces
      %w[
        s:p
        s:l
        s:n
        s:s
        s:b
        s:r
        S:P
        S:L
        S:N
        S:S
        S:B
        S:R
      ]
    end

    def promotable?(piece)
      promotable_pieces.include?(piece)
    end

    def promote(piece)
      return piece unless promotable?(piece)

      piece.gsub(/:\+?/, ":+")
    end

    def unpromote(piece)
      piece.gsub(/:\+?/, ":")
    end

    def owned(piece)
      raise Error::GameOver::OpponentPiece unless owned?(piece)

      piece
    end

    def owned?(piece)
      if last_position.turn_to_north?
        owned_by_north?(piece)
      else
        owned_by_south?(piece)
      end
    end

    def owned_by_north?(piece)
      piece.downcase == piece
    end

    def owned_by_south?(piece)
      piece.upcase == piece
    end

    def own(piece)
      if last_position.turn_to_north?
        piece.downcase
      else
        piece.upcase
      end
    end

    def time_forfeit?(timestamp)
      seconds_per_move(timestamp) > @max_seconds_per_move
    end

    def last_state
      @states.fetch(-1)
    end

    def last_position
      last_state.position
    end

    def last_timestamp
      last_state.timestamp
    end

    def seconds_per_move(timestamp)
      timestamp - last_timestamp
    end
  end
end
