# frozen_string_literal: true

require "qi"

require_relative "error/game_over/illegal_move/piece_not_found"

module Nichugi
  # The position abstraction.
  class Position
    SHOGI_VS_XIANGQI99_SQUARE = {
      0  => "x:r",
      1  => "x:h",
      2  => "x:e",
      3  => "x:a",
      4  => "x:-g",
      5  => "x:a",
      6  => "x:e",
      7  => "x:h",
      8  => "x:r",
      19 => "x:c",
      25 => "x:c",
      27 => "x:s",
      29 => "x:s",
      31 => "x:s",
      33 => "x:s",
      35 => "x:s",
      54 => "S:P",
      55 => "S:P",
      56 => "S:P",
      57 => "S:P",
      58 => "S:P",
      59 => "S:P",
      60 => "S:P",
      61 => "S:P",
      62 => "S:P",
      64 => "S:B",
      70 => "S:R",
      72 => "S:L",
      73 => "S:N",
      74 => "S:S",
      75 => "S:G",
      76 => "S:-K",
      77 => "S:G",
      78 => "S:S",
      79 => "S:N",
      80 => "S:L"
    }.freeze

    XIANGQI99_VS_SHOGI_SQUARE = {
      0  => "s:l",
      1  => "s:n",
      2  => "s:s",
      3  => "s:g",
      4  => "s:-k",
      5  => "s:g",
      6  => "s:s",
      7  => "s:n",
      8  => "s:l",
      10 => "s:r",
      16 => "s:b",
      18 => "s:p",
      19 => "s:p",
      20 => "s:p",
      21 => "s:p",
      22 => "s:p",
      23 => "s:p",
      24 => "s:p",
      25 => "s:p",
      26 => "s:p",

      45 => "X:S",
      47 => "X:S",
      49 => "X:S",
      51 => "X:S",
      53 => "X:S",
      55 => "X:C",
      61 => "X:C",
      72 => "X:R",
      73 => "X:H",
      74 => "X:E",
      75 => "X:A",
      76 => "X:-G",
      77 => "X:A",
      78 => "X:E",
      79 => "X:H",
      80 => "X:R"
    }.freeze

    def self.shogi_vs_xiangqi99
      new(in_hand: [], turn_to_north: false, square: SHOGI_VS_XIANGQI99_SQUARE)
    end

    def self.xiangqi99_vs_shogi
      new(in_hand: [], turn_to_north: false, square: XIANGQI99_VS_SHOGI_SQUARE)
    end

    def self.squares_count
      81
    end

    def self.drop?(square)
      square >= squares_count
    end

    def self.drop_index(square)
      square % squares_count
    end

    attr_reader :square

    # Initialize a position.
    #
    # @param in_hand [Array] The list of pieces in hand.
    # @param turn_to_north [Boolean] The identifier of the player who must play.
    # @param square [Hash] The index of each piece on the board.
    #
    # @example Dump a classic Tsume Shogi problem
    #   new(
    #     "in_hand": [S r r b g g g g s n n n n p p p p p p p p p p p p p p p p p],
    #     "turn_to_north": false,
    #     "square": {
    #        3 => "s",
    #        4 => "k",
    #        5 => "s",
    #       22 => "+P",
    #       43 => "+B"
    #     }
    #   )
    def initialize(in_hand:, turn_to_north:, square:)
      @in_hand = in_hand
      @turn_to_north = turn_to_north
      @square = square
    end

    def after(source, target, new_piece, captured_piece = nil)
      hash = ::Qi.call(source, target, new_piece, captured_piece, square: @square, in_hand: @in_hand)
      self.class.new(turn_to_north: turn_to_south?, **hash)
    end

    def piece(index, from_hand:)
      pieces(from_hand: from_hand).fetch(index) { raise Error::GameOver::IllegalMove::PieceNotFound }
    end

    def pieces(from_hand:)
      if from_hand
        @in_hand
      else
        @square
      end
    end

    def promotion_square?(index)
      promotion_square_ids.include?(index)
    end

    # @return [Boolean] Is turn to north?
    def turn_to_north?
      @turn_to_north
    end

    # @return [Boolean] Is turn to south?
    def turn_to_south?
      !turn_to_north?
    end

    # @param conditions [Hash] A hash of integers and symbols.
    #
    # @example A hash of square patterns
    #   { 10 => nil, 20 => nil, 30 => nil, 40 => nil, 50 => nil }
    #
    # @return [Boolean] Is the position matching the given pattern?
    def match?(conditions)
      conditions.all? { |index, state| correct?(index, state) }
    end

    # @param index [Integer] The square ID.
    #
    # @return [String, nil] The square content.
    def square_content(index)
      @square.fetch(index, nil)
    end

    # @note Faster than FEEN to compute.
    def inspect
      inspect_square = @square.keys.sort.inject("") { |s, key| s + "#{key}:#{@square.fetch(key)}" }
      inspect_turn = (turn_to_north? ? "north-turn" : "south-turn")
      inspect_in_hand = @in_hand.sort.join(";")

      "<Nichugi::Position #{inspect_square} #{inspect_turn} #{inspect_in_hand}>"
    end

    protected

    # @param square_id [Integer] The square ID.
    # @param state [Symbol] The state of the square.
    #
    # @return [Boolean] Is the state of the square correct?
    def correct?(square_id, state)
      content = square_content(square_id)

      case state
      when :occupied
        !content.nil?
      when :enemy
        if turn_to_north?
          !content.nil? && content.upcase == content
        else
          !content.nil? && content.downcase == content
        end
      else
        state == content
      end
    end

    def promotion_square_ids
      if turn_to_north?
        north_promotion_square_ids
      else
        south_promotion_square_ids
      end
    end

    def north_promotion_square_ids
      (54..80)
    end

    def south_promotion_square_ids
      (0..26)
    end
  end
end
